/*
 * Copy me, if you can.
 * by 20h
 */

#ifdef linux
#define _XOPEN_SOURCE 500
#endif

#define _FILE_OFFSET_BITS 64
#define FUSE_USE_VERSION 25
#include <fuse.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include "libixp/ixp.h"
#include "libcext/cext.h"

static IXPClient c = { 0 };

static mode_t
mkunixmode(unsigned int mode)
{
	mode_t umode;

	umode = 0;

	if(mode & IXP_DMDIR)
		umode |= S_IFDIR;
	else
		umode |= S_IFREG;
	umode |= (mode & 0x00000777);

	return umode;
}

static int
ninep_getattr(const char *path, struct stat *stbuf)
{
	unsigned int fid;
	Stat s;
	unsigned char *buf;

	memset(&s, 0, sizeof(Stat));
	fid = c.root_fid << 2;
	if(ixp_client_stat(&c, fid, (char *)path) < 0)
		return -1;

	buf = c.ofcall.stat;
	ixp_unpack_stat(&buf, nil, &s);
	stbuf->st_mode = mkunixmode(s.mode);
	stbuf->st_nlink = 1;
	stbuf->st_uid = 0;
	stbuf->st_gid = 0;
	stbuf->st_rdev = 0;
	stbuf->st_size = s.length;
	stbuf->st_blocks = 0;
	stbuf->st_atime = s.atime;
	stbuf->st_mtime = s.mtime;
	stbuf->st_ctime = 0;

	ixp_client_close(&c, fid);
	return 0;
}

static int
ninep_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                       off_t offset, struct fuse_file_info *fi)
{
	unsigned int fid;
	Stat s;
	unsigned char *buffer;
	int count;
	unsigned long long off;
	void *data;
	static unsigned char result[IXP_MAX_MSG];
	struct stat st;

	fid = c.root_fid << 2;
	off = 0;
	data = nil;

	if(ixp_client_walkopen(&c, fid, (char *)path, IXP_OREAD) < 0)
                return -ENOENT;
	while((count = ixp_client_read(&c, fid, off, result, IXP_MAX_MSG)) > 0) {
		data = cext_erealloc(data, off + count);
		memmove(data + off, result, count);
	       	off += count;
	}
	if(count < 0)
		return -1;

	filler(buf, ".", nil, 0);
	filler(buf, "..", nil, 0);
	if(off == 0)
		goto goodend;

	buffer = data;
	do {
		ixp_unpack_stat(&buffer, nil, &s);
		memset(&st, 0, sizeof(st));
		st.st_mode = mkunixmode(s.mode);
		if(filler(buf, s.name, &st, 0))
			break;
	} while(((int)buffer - (int)data) < off + count); 

goodend:
	ixp_client_close(&c, fid);
	
	return 0;
}

static int
ninep_mknod(const char *path, mode_t mode, dev_t rdev)
{
	unsigned int fid;
	char *p;

	p = strrchr(path, '/');
	if(p == NULL)
		return -1;
	*p = '\0';
	if(ixp_client_walk(&c, fid, (char *)path) < 0)
		return -1;
	p++;
	if(ixp_client_create(&c, fid, p, mode, mode) < 0)
		return -1;
	ixp_client_close(&c, fid);

	return 0;
}

static int
ninep_mkdir(const char *path, mode_t mode)
{
	unsigned int fid;
	char *p;

	p = strrchr(path, '/');
	if(p == nil)
		return -1;
	*p = '\0';
	if(ixp_client_walk(&c, fid, (char *)path) < 0)
		return -1;
	p++;
	if(ixp_client_create(&c, fid, p, mode, mode) < 0)
		return -1;
	ixp_client_close(&c, fid);

	return 0;
}

static int
ninep_unlink(const char *path)
{
	unsigned int fid;

	fid = c.root_fid << 2;
	if(ixp_client_remove(&c, fid, (char *)path) < 0)
		return -1;

	return 0;
}

static int
ninep_rmdir(const char *path)
{
	unsigned int fid;

	fid = c.root_fid << 2; 
	ixp_client_remove(&c, fid, (char *)path);

	return 0;
}

static int
ninep_open(const char *path, struct fuse_file_info *fi)
{
    return 0;
}

static int
ninep_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
	unsigned int fid;
	int res, count, read;
	static unsigned char result[IXP_MAX_MSG];

	fid = c.root_fid << 2;
	res = 0;
	(void) fi;

	if(ixp_client_walkopen(&c, fid, (char *)path, IXP_OREAD) == -1)
		return -ENOENT;

	if(size > IXP_MAX_MSG)
		read = IXP_MAX_MSG;
	else
		read = size;
	while((count = ixp_client_read(&c, fid, offset + res, result, read)) > 0) {
		fprintf(stderr, "size: %d\n", size);
		memmove(buf + res, result, count);
		res += count;	
		size -= read;
		if(size > IXP_MAX_MSG)
			read = IXP_MAX_MSG;
		else
			read = size;
	}

	if(count < 0)
		return -1;
	ixp_client_close(&c, fid);

	return res;
}

static int
ninep_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
	unsigned int fid;
	int res;

	fid = c.root_fid << 2;
	res = size;
	(void) fi;

	if(ixp_client_walkopen(&c, fid, (char *)path, IXP_OWRITE) == -1)
		return -1;

	if(ixp_client_write(&c, fid, offset,
			     	size, (unsigned char *)buf) != size)
		res  = -1;
	ixp_client_close(&c, fid);
		
	return res;
}

static struct fuse_operations ninep_oper = {
	.getattr	= ninep_getattr,
	.readdir	= ninep_readdir,
	.mknod		= ninep_mknod,
	.mkdir		= ninep_mkdir,
	.unlink		= ninep_unlink,
	.rmdir		= ninep_rmdir,
	.open		= ninep_open,
	.read		= ninep_read,
	.write		= ninep_write,
};

int
main(int argc, char *argv[])
{
	
	if(argc > 2) {
		if(ixp_client_dial(&c, argv[1], getpid()) < 0) {
			fprintf(stderr, "connect error: %s\n", c.errstr);
			exit(1);
		}
		argv++;
		argc--;
	}

	umask(0);
	return fuse_main(argc, argv, &ninep_oper);
}
